import { Renderer } from "./renderer.ts";

export class V2 {
    constructor(
        public x: number,
        public y: number,
    ) {}

    clone(): V2 {
        return new V2(this.x, this.y);
    }

    add(other: V2): V2 {
        return new V2(this.x + other.x, this.y + other.y);
    }

    sub(other: V2): V2 {
        return new V2(this.x - other.x, this.y - other.y);
    }

    mul(factor: number): V2 {
        return new V2(this.x * factor, this.y * factor);
    }

    div(factor: number): V2 {
        return new V2(this.x / factor, this.y / factor);
    }

    pow(factor: number): V2 {
        return new V2(this.x ** factor, this.y ** factor);
    }

    sum(): number {
        return this.x + this.y;
    }

    len(): number {
        return Math.sqrt(this.pow(2).sum());
    }

    abs(): V2 {
        return new V2(Math.abs(this.x), Math.abs(this.y));
    }
}

export interface Geometry {
    pointInside(thisPos: V2, pointPos: V2): boolean;
    collidesWith(thisPos: V2, other: Geometry, otherPos: V2): boolean;
    boundingRectSize(): V2;
    render(r: Renderer, pos: V2, color: string): void;
}

export class Rect implements Geometry {
    constructor(
        public width: number,
        public height: number,
    ) {}

    v2(): V2 {
        return new V2(this.width, this.height);
    }

    pointInside(thisPos: V2, pointPos: V2): boolean {
        const { x, y } = thisPos;
        const { width: w, height: h } = this;
        const { x: ox, y: oy } = pointPos;
        return ox > x && ox < x + w && oy > y && oy < y + h;
    }

    collidesWith(thisPos: V2, other: Geometry, otherPos: V2): boolean {
        if (other instanceof Rect) {
            const { x, y } = thisPos;
            const { width: w, height: h } = this;
            const { x: ox, y: oy } = otherPos;
            const { width: ow, height: oh } = other;
            return ox + ow > x && ox < x + w && oy + oh > y && oy < y + h;
        }
        return other.collidesWith(otherPos, this, thisPos);
    }

    boundingRectSize(): V2 {
        return new V2(this.width, this.height);
    }

    render(r: Renderer, pos: V2, color: string): void {
        r.fillRect(pos.x, pos.y, this.width, this.height, color);
    }
}

export class Circle implements Geometry {
    constructor(
        public radius: number,
    ) {}

    pointInside(thisPos: V2, pointPos: V2): boolean {
        return pointPos.sub(thisPos).len() < this.radius;
    }

    collidesWith(thisPos: V2, other: Geometry, otherPos: V2): boolean {
        if (other instanceof Circle) {
            return otherPos.sub(thisPos).len() < other.radius;
        }
        if (other instanceof Rect) {
            const circleDistance = thisPos.sub(otherPos).abs();
            const halfRect = other.v2().div(2);
            if (
                circleDistance.x >= thisPos.x + halfRect.x ||
                circleDistance.y >= thisPos.y + halfRect.y
            ) {
                return false;
            }
            if (
                circleDistance.x < halfRect.x || circleDistance.y < halfRect.y
            ) {
                return true;
            }
            return circleDistance.add(halfRect).pow(2).sum() < this.radius ** 2;
        }
        return other.collidesWith(otherPos, other, thisPos);
    }

    boundingRectSize(): V2 {
        return new V2(this.radius, this.radius);
    }

    render(r: Renderer, pos: V2, color: string): void {
        r.fillCirc(pos.x, pos.y, this.radius, color);
    }
}

export class Shape implements Geometry {
    constructor(
        public innerShapes: [V2, Geometry][],
    ) {}

    pointInside(thisPos: V2, pointPos: V2): boolean {
        return this.innerShapes
            .some(([pos, shape]) =>
                shape.pointInside(thisPos.add(pos), pointPos)
            );
    }

    collidesWith(thisPos: V2, other: Geometry, otherPos: V2): boolean {
        return this.innerShapes
            .some(([pos, shape]) =>
                other.collidesWith(thisPos.add(pos), shape, otherPos)
            );
    }

    boundingRectSize(): V2 {
        return this.innerShapes
            .reduce((size, [pos, shape]) => {
                const shapeSize = shape.boundingRectSize();
                size.x = Math.max(size.x, pos.x + shapeSize.x);
                size.y = Math.max(size.y, pos.y + shapeSize.y);
                return size;
            }, new V2(0, 0));
    }

    render(r: Renderer, pos: V2, color: string): void {
        for (const [shapePos, shape] of this.innerShapes) {
            shape.render(r, shapePos.add(pos), color);
        }
    }
}
