import { EvHandler, Mouse } from "./input.ts";
import { Renderer, RendererImage } from "./renderer.ts";

export class Grid {
    private t: Transformation = { s: 2, ox: 0, oy: 0 };
    private transformer = new Transformer(this.t);
    private tr = new TransformingRenderer(this.t);

    private _transformingMouse: TransformingMouse;

    constructor(
        private mouse: Mouse,
    ) {
        this._transformingMouse = new TransformingMouse(this.mouse, this.t);

        this.mouse.addOnPress(() => {
            this.transformer.startPan();
            return "stop";
        });
        this.mouse.addOnRelease(() => {
            this.transformer.endPan();
            return "stop";
        });
        this.mouse.addOnMove(() => {
            const { x, y } = this.mouse;
            this.transformer.move(x, y);
            return "stop";
        });
        this.mouse.addOnScroll((direction) => {
            if (direction === "up") {
                this.transformer.zoomIn();
            } else {
                this.transformer.zoomOut();
            }
            return "stop";
        });
    }

    transformingMouse(): Mouse {
        return this._transformingMouse;
    }

    render(r: Renderer, renderTransformed: (r: Renderer) => void) {
        this.transformer.updateCanvas(r.width, r.height);
        this.drawGrid(r);
        this.tr.render(r, (tr) => {
            renderTransformed(tr);
        });
    }

    private drawGrid(r: Renderer) {
        const { t: { s, ox, oy } } = this;
        r.clear("#ddd");

        const dotSpace = 32 * s;

        const dotOffsetX = ox % dotSpace - dotSpace;
        const dotOffsetY = oy % dotSpace - dotSpace;

        const dotWidth = r.width / dotSpace + 1;
        const dotHeight = r.height / dotSpace + 1;

        if (s < 0.2) {
            return;
        }
        for (let y = 0; y < Math.min(dotHeight, 200); ++y) {
            for (let x = 0; x < Math.min(dotWidth, 200); ++x) {
                r.fillCirc(
                    x * dotSpace + dotOffsetX + 16 * s,
                    y * dotSpace + dotOffsetY + 16 * s,
                    2 * s,
                    "#bbb",
                );
            }
        }
    }
}

// offset is screen offset
type Transformation = {
    s: number;
    ox: number;
    oy: number;
};

class Transformer {
    private isPanning = false;

    private lastMouseX = 0;
    private lastMouseY = 0;
    private lastCanvasWidth = 0;
    private lastCanvasHeight = 0;

    constructor(
        private t: Transformation,
    ) {}

    updateCanvas(width: number, height: number) {
        this.lastCanvasWidth = width;
        this.lastCanvasHeight = height;
    }

    startPan() {
        this.isPanning = true;
    }

    endPan() {
        this.isPanning = false;
    }

    move(x: number, y: number) {
        const deltaX = x - this.lastMouseX;
        const deltaY = y - this.lastMouseY;
        this.lastMouseX = x;
        this.lastMouseY = y;
        if (this.isPanning) {
            this.t.ox += deltaX;
            this.t.oy += deltaY;
        }
    }

    zoomIn() {
        this.zoom(1 - 1 / 1.1);
    }
    zoomOut() {
        this.zoom(1 - 1.1);
    }

    private zoom(factor: number) {
        this.t.s *= 1 + factor;

        const mouseRatioX = (this.lastMouseX - this.t.ox) /
            this.lastCanvasWidth;
        const mouseRatioY = (this.lastMouseY - this.t.oy) /
            this.lastCanvasHeight;

        const deltaOffsetX = this.lastCanvasWidth * factor *
            mouseRatioX * -1;
        const deltaOffsetY = this.lastCanvasHeight * factor *
            mouseRatioY * -1;

        this.t.ox += deltaOffsetX;
        this.t.oy += deltaOffsetY;
    }
}

class TransformingRenderer implements Renderer {
    private r!: Renderer;

    constructor(
        private t: Transformation,
    ) {}

    render(r: Renderer, renderTransformed: (r: Renderer) => void): void {
        this.r = r;
        renderTransformed(this);
    }

    get width(): number {
        return this.r.width / this.t.s;
    }
    get height(): number {
        return this.r.height / this.t.s;
    }
    clear(color: string): void {
        this.r.clear(color);
    }
    fillRect(x: number, y: number, w: number, h: number, color: string): void {
        const { r, t: { s, ox, oy } } = this;
        r.fillRect(x * s + ox, y * s + oy, w * s, h * s, color);
    }
    strokeRect(
        x: number,
        y: number,
        w: number,
        h: number,
        color: string,
        lineWidth: number,
    ): void {
        const { r, t: { s, ox, oy } } = this;
        r.strokeRect(
            x * s + ox,
            y * s + oy,
            w * s,
            h * s,
            color,
            lineWidth * s,
        );
    }
    fillCirc(x: number, y: number, radius: number, color: string): void {
        const { r, t: { s, ox, oy } } = this;
        r.fillCirc(x * s + ox, y * s + oy, radius * s, color);
    }
    strokeLine(
        x0: number,
        y0: number,
        x1: number,
        y1: number,
        color: string,
        lineWidth: number,
    ): void {
        const { r, t: { s, ox, oy } } = this;
        r.strokeLine(
            x0 * s + ox,
            y0 * s + oy,
            x1 * s + ox,
            y1 * s + oy,
            color,
            lineWidth * s,
        );
    }
    putImage(
        data: RendererImage,
        x: number,
        y: number,
        w = data.width,
        h = data.width,
        alpha = 1,
    ): void {
        const { r, t: { s, ox, oy } } = this;

        r.putImage(
            data,
            x * s + ox,
            y * s + oy,
            w * s,
            h * s,
            alpha,
        );
    }
}

class TransformingMouse implements Mouse {
    constructor(
        private mouse: Mouse,
        private t: Transformation,
    ) {}

    get x(): number {
        return (this.mouse.x - this.t.ox) / this.t.s;
    }
    get y(): number {
        return (this.mouse.y - this.t.oy) / this.t.s;
    }
    addOnPress(handler: EvHandler): void {
        this.mouse.addOnPress(handler);
    }
    addOnRelease(handler: EvHandler): void {
        this.mouse.addOnRelease(handler);
    }
    addOnMove(handler: EvHandler): void {
        this.mouse.addOnMove(handler);
    }
    addOnScroll(handler: EvHandler<["up" | "down"]>): void {
        this.mouse.addOnScroll(handler);
    }
}
