import { Circle, Geometry, Rect, V2 } from "./geometry.ts";
import { Grid } from "./grid.ts";
import { Mouse } from "./input.ts";
import { Painter } from "./painter.ts";
import { Renderer } from "./renderer.ts";

export class Simulator {
    private grid: Grid;
    private circuit: Circuit;
    private tooltip: Tooltip;
    private toolbar: Toolbar;

    constructor(
        private mouse: Mouse,
    ) {
        this.grid = new Grid(this.mouse);
        this.circuit = new Circuit(this.grid.transformingMouse());
        this.tooltip = new Tooltip(this.circuit, this.grid.transformingMouse());
        this.toolbar = new Toolbar(this.mouse, this.tooltip);
    }

    render(r: Renderer) {
        hover: {
            if (this.toolbar.hover() === "break") {
                break hover;
            }
            if (this.tooltip.hover() === "break") {
                break hover;
            }
            document.body.style.cursor = "default";
        }

        r.clear("black");
        this.grid.render(r, (tr) => {
            this.circuit.render(tr);
            this.toolbar.render(r);
            this.tooltip.render(tr);
        });
    }
}

type Tool = {
    component: Component;
    offsetX: number;
};

class Toolbar {
    private tools: Tool[] = [
        { component: new SwitchComponent(), offsetX: 32 },
        { component: new LedComponent(), offsetX: 256 },
    ];

    private lastWidth = 0;
    private lastHeight = 0;

    private hoveringComponentIdx?: number;

    private scale = 64;

    constructor(
        private mouse: Mouse,
        private tooltip: Tooltip,
    ) {
        this.mouse.addOnPress(() => {
            const { x, y } = this.mouse;

            if (
                !(x >= 0 && y >= this.lastHeight - 200 && x < this.lastWidth &&
                    y < this.lastHeight)
            ) {
                return "bubble";
            }

            for (const [i, tool] of this.tools.entries()) {
                const compWidth = tool.component.width * this.scale;
                const compHeight = tool.component.height * this.scale;
                if (
                    x >= tool.offsetX &&
                    y >= this.lastHeight - 100 - 48 &&
                    x < tool.offsetX + compWidth &&
                    y < this.lastHeight - 100 + compHeight
                ) {
                    this.tooltip.select(this.tools[i].component);
                    return "stop";
                }
            }
            this.tooltip.deselect();
            return "stop";
        });
    }

    hover(): "continue" | "break" {
        const { x, y } = this.mouse;
        for (const [i, tool] of this.tools.entries()) {
            const compWidth = tool.component.width * this.scale;
            const compHeight = tool.component.height * this.scale;
            if (
                x >= tool.offsetX &&
                y >= this.lastHeight - 100 &&
                x < tool.offsetX + compWidth &&
                y < this.lastHeight - 100 + compHeight
            ) {
                this.hoveringComponentIdx = i;
                document.body.style.cursor = "pointer";
                return "break";
            }
        }
        this.hoveringComponentIdx = undefined;
        return "continue";
    }

    render(r: Renderer): void {
        this.lastWidth = r.width;
        this.lastHeight = r.height;

        r.fillRect(0, r.height - 200, r.width, 200, "#aaa");

        for (const [i, tool] of this.tools.entries()) {
            const compWidth = tool.component.width * this.scale;
            const compHeight = tool.component.height * this.scale;
            r.strokeRect(
                tool.offsetX,
                r.height - 100,
                compWidth,
                compHeight,
                "#777",
                1,
            );
            if (i === this.hoveringComponentIdx) {
                r.fillRect(
                    tool.offsetX,
                    r.height - 100,
                    compWidth,
                    compHeight,
                    "#00000088",
                );
            }
            tool.component.render(r, tool.offsetX, r.height - 100);
        }
    }
}

class Tooltip {
    private selectedComponent?: Component;

    private shouldHover = false;

    private isWiring = false;
    private selectedOutputTerm?: ComponentWireTerm;
    private selectedInputTerm?: ComponentWireTerm;

    constructor(
        private circuit: Circuit,
        public mouse: Mouse,
    ) {
        this.mouse.addOnPress(() => {
            if (this.selectedComponent) {
                const x = Math.floor(this.mouse.x / 32);
                const y = Math.floor(this.mouse.y / 32);
                this.circuit.tryPlace(this.selectedComponent.clone(), x, y);
                return "stop";
            } else if (!this.isWiring && this.selectedOutputTerm) {
                this.isWiring = true;
                return "stop";
            } else if (this.isWiring && this.selectedInputTerm) {
                if (!this.selectedOutputTerm) {
                    throw new Error("invalid state");
                }
                this.circuit.wire(
                    this.selectedOutputTerm,
                    this.selectedInputTerm,
                );
                this.isWiring = false;
                this.selectedOutputTerm = undefined;
                this.selectedInputTerm = undefined;
            } else {
                this.isWiring = false;
                this.selectedOutputTerm = undefined;
            }

            return "bubble";
        });
    }

    hover(): "continue" | "break" {
        if (this.selectedComponent) {
            const x = Math.floor(this.mouse.x / 32);
            const y = Math.floor(this.mouse.y / 32);

            this.shouldHover = false;
            if (!this.circuit.placeIsOccupied(x, y)) {
                this.shouldHover = true;
                return "break";
            }
        } else if (!this.isWiring) {
            const term = this.circuit.hoveredOutputTerminal();
            this.selectedOutputTerm = undefined;
            if (term) {
                this.selectedOutputTerm = term;
                document.body.style.cursor = "pointer";
                return "break";
            }
        } else {
            const term = this.circuit.hoveredInputTerminal();
            this.selectedInputTerm = undefined;
            if (term) {
                this.selectedInputTerm = term;
                document.body.style.cursor = "pointer";
                return "break";
            }
        }

        return "continue";
    }

    render(r: Renderer): void {
        if (this.selectedComponent) {
            const x = Math.floor(this.mouse.x / 32) * 32;
            const y = Math.floor(this.mouse.y / 32) * 32;
            if (this.shouldHover) {
                this.selectedComponent
                    .renderTransparent(
                        r,
                        x - this.selectedComponent.offsetX * 32,
                        y - this.selectedComponent.height * 32,
                    );
            }
        } else if (!this.isWiring) {
            const output = this.selectedOutputTerm;
            if (output) {
                output.geometry.render(r, output.pos, "#777");
            }
        } else if (this.isWiring) {
            const output = this.selectedOutputTerm;
            if (!output) {
                throw new Error("invalid state");
            }
            const input = this.selectedInputTerm;
            if (input) {
                input.geometry.render(r, input.pos, "#777");
                r.strokeLine(
                    output.pos.x,
                    output.pos.y,
                    input.pos.x,
                    input.pos.y,
                    "222",
                    2,
                );
            } else {
                r.strokeLine(
                    output.pos.x,
                    output.pos.y,
                    this.mouse.x,
                    this.mouse.y,
                    "222",
                    2,
                );
            }
        }
    }

    select(component: Component): void {
        this.selectedComponent = component;
    }

    deselect(): void {
        this.selectedComponent = undefined;
    }
}

type PlacedComponent = {
    component: Component;
    x: number;
    y: number;
};

interface WireSource {
    get x(): number;
    get y(): number;
    high(): boolean;
}

interface WireDest {
    get x(): number;
    get y(): number;
    setHigh(): void;
    setLow(): void;
}

type Wire = {
    source: WireSource;
    dest: WireDest;
};

class Circuit {
    private components: PlacedComponent[] = [];
    private wires: Wire[] = [];

    constructor(
        private mouse: Mouse,
    ) {
        this.mouse.addOnPress(() => {
            for (const { component, x, y } of this.components) {
                if (!component.click) {
                    continue;
                }
                if (
                    this.mouse.x >= x * 32 &&
                    this.mouse.y >= y * 32 &&
                    this.mouse.x < x * 32 + component.width * 32 &&
                    this.mouse.y < y * 32 + component.height * 32
                ) {
                    component.click(
                        this.mouse.x - x * 32,
                        this.mouse.y - y * 32,
                    );
                    return "stop";
                }
            }
            return "bubble";
        });
    }

    placeIsOccupied(x: number, y: number): boolean {
        return this.components.some((c) =>
            c.component.collidesWith(
                new V2(c.x, c.y),
                c.component,
                new V2(x, y),
            )
        );
    }

    hoveredOutputTerminal(): ComponentWireTerm | null {
        for (const { x, y, component } of this.components) {
            const term = component.hoveredOutputTerminal(
                new V2(x, y).mul(32),
                new V2(this.mouse.x, this.mouse.y),
            );
            if (term !== null) {
                return term;
            }
        }
        return null;
    }

    hoveredInputTerminal(): ComponentWireTerm | null {
        for (const { x, y, component } of this.components) {
            const term = component.hoveredInputTerminal(
                new V2(x, y).mul(32),
                new V2(this.mouse.x, this.mouse.y),
            );
            if (term !== null) {
                return term;
            }
        }
        return null;
    }

    tryPlace(component: Component, x: number, y: number): void {
        if (this.placeIsOccupied(x, y)) {
            return;
        }
        this.components.push({ component, x, y });
    }

    wire(source: WireSource, dest: WireDest): void {
        this.wires.push({ source, dest });
    }

    render(r: Renderer): void {
        for (const { source, dest } of this.wires) {
            r.strokeLine(source.x, source.y, dest.x, dest.y, "black", 3);
        }
        for (const { component, x, y } of this.components) {
            component.render(
                r,
                x * 32 - component.offsetX * 32,
                y * 32 - component.offsetY * 32,
            );
        }
    }
}

class ComponentWireTerm implements WireSource, WireDest {
    constructor(
        public pos: V2,
        public geometry: Geometry,
        public component: Component,
    ) {}
    get x(): number {
        return this.pos.x;
    }
    get y(): number {
        return this.pos.y;
    }
    high(): boolean {
        throw new Error("Method not implemented.");
    }
    setHigh(): void {
        throw new Error("Method not implemented.");
    }
    setLow(): void {
        throw new Error("Method not implemented.");
    }
}

interface Component {
    get offsetX(): number;
    get offsetY(): number;
    get width(): number;
    get height(): number;

    collidesWith(thisPos: V2, other: Component, otherPos: V2): boolean;
    collidesWithGeometry(thisPos: V2, other: Geometry, otherPos: V2): boolean;
    render(r: Renderer, x: number, y: number): void;
    renderTransparent(r: Renderer, x: number, y: number): void;
    click?(x: number, y: number): void;
    hoveredInputTerminal(pos: V2, mousePos: V2): ComponentWireTerm | null;
    hoveredOutputTerminal(pos: V2, mousePos: V2): ComponentWireTerm | null;

    clone(): Component;
}

class SwitchComponent implements Component {
    private graphicOn = new Painter(96, 64);
    private graphicOff = new Painter(96, 64);

    private switchOn = false;

    private geometry = new Rect(1.5, 1);

    constructor() {
        this.graphicOff.strokeRect(
            0,
            0,
            this.graphicOff.width,
            this.graphicOff.height,
            "#88000088",
            2,
        );
        this.graphicOff.fillRect(48, 30, 16, 4, "black");
        this.graphicOff.fillCirc(64, 32, 8, "black");
        this.graphicOff.fillCirc(32, 32, 16, "black");

        this.graphicOn.strokeRect(
            0,
            0,
            this.graphicOn.width,
            this.graphicOn.height,
            "#88000088",
            2,
        );
        this.graphicOn.fillRect(48, 30, 16, 4, "black");
        this.graphicOn.fillCirc(64, 32, 8, "black");
        this.graphicOn.fillCirc(32, 32, 16, "red");
    }

    get offsetX(): number {
        return 0.5;
    }
    get offsetY(): number {
        return 0.5;
    }
    get width(): number {
        return this.geometry.width;
    }
    get height(): number {
        return this.geometry.height;
    }

    clone(): Component {
        return new SwitchComponent();
    }

    collidesWith(thisPos: V2, other: Component, otherPos: V2): boolean {
        return other.collidesWithGeometry(otherPos, this.geometry, thisPos);
    }

    collidesWithGeometry(thisPos: V2, other: Geometry, otherPos: V2): boolean {
        return this.geometry.collidesWith(thisPos, other, otherPos);
    }

    render(r: Renderer, x: number, y: number): void {
        const graphic = this.switchOn ? this.graphicOn : this.graphicOff;
        graphic.render(r, x, y);
    }

    renderTransparent(r: Renderer, x: number, y: number): void {
        const graphic = this.switchOn ? this.graphicOn : this.graphicOff;
        graphic.render(r, x, y, 0.5);
    }

    click(x: number, y: number): void {
        const onButton = Math.sqrt(
            (x - 16) ** 2 + (y - 16) ** 2,
        ) <= 16;

        if (onButton) {
            this.switchOn = !this.switchOn;
        }
    }

    hoveredInputTerminal(
        _componentPos: V2,
        _mousePos: V2,
    ): ComponentWireTerm | null {
        return null;
    }

    hoveredOutputTerminal(
        componentPos: V2,
        mousePos: V2,
    ): ComponentWireTerm | null {
        const geometry = new Circle(8);
        const pos = new V2(48, 16).add(componentPos);

        if (geometry.pointInside(pos, mousePos)) {
            return new ComponentWireTerm(pos, geometry, this);
        }
        return null;
    }
}

class LedComponent implements Component {
    private graphicOn = new Painter(96, 64);
    private graphicOff = new Painter(96, 64);

    private switchOn = false;

    private geometry = new Rect(1.5, 1);

    constructor() {
        this.graphicOff.strokeRect(
            0,
            0,
            this.graphicOff.width,
            this.graphicOff.height,
            "#88000088",
            2,
        );
        this.graphicOff.fillRect(32, 30, 16, 4, "black");
        this.graphicOff.fillCirc(32, 32, 8, "black");
        this.graphicOff.fillCirc(64, 32, 16, "black");

        this.graphicOn.strokeRect(
            0,
            0,
            this.graphicOn.width,
            this.graphicOn.height,
            "#88000088",
            2,
        );
        this.graphicOn.fillRect(32, 30, 16, 4, "black");
        this.graphicOn.fillCirc(32, 32, 8, "black");
        this.graphicOn.fillCirc(64, 32, 16, "red");
    }

    get offsetX(): number {
        return 1.5;
    }
    get offsetY(): number {
        return 0.5;
    }
    get width(): number {
        return this.geometry.width;
    }
    get height(): number {
        return this.geometry.height;
    }

    clone(): Component {
        return new LedComponent();
    }

    collidesWith(thisPos: V2, other: Component, otherPos: V2): boolean {
        return other.collidesWithGeometry(
            otherPos,
            this.geometry,
            thisPos.sub(new V2(0, 0)),
        );
    }

    collidesWithGeometry(thisPos: V2, other: Geometry, otherPos: V2): boolean {
        return this.geometry.collidesWith(thisPos, other, otherPos);
    }

    render(r: Renderer, x: number, y: number): void {
        const graphic = this.switchOn ? this.graphicOn : this.graphicOff;
        graphic.render(r, x, y);
    }

    renderTransparent(r: Renderer, x: number, y: number): void {
        const graphic = this.switchOn ? this.graphicOn : this.graphicOff;
        graphic.render(r, x, y, 0.5);
    }

    hoveredInputTerminal(
        componentPos: V2,
        mousePos: V2,
    ): ComponentWireTerm | null {
        const geometry = new Circle(8);
        const pos = new V2(-16, 16).add(componentPos);

        if (geometry.pointInside(pos, mousePos)) {
            return new ComponentWireTerm(pos, geometry, this);
        }
        return null;
    }

    hoveredOutputTerminal(
        _componentPos: V2,
        _mousePos: V2,
    ): ComponentWireTerm | null {
        return null;
    }
}
